"""Implement semantic entropy."""
import os
import pickle
import logging

import numpy as np
import torch
import torch.nn.functional as F

# from transformers import AutoModelForSequenceClassification, AutoTokenizer

# from .pretrained_models.huggingface_models import HuggingfaceModel
from ...utils.az import AzureModel
from .utils import utils

# os.environ['USE_MPS'] = 'true'  # for M3

DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
# DEVICE = torch.device("mps") if os.environ.get('USE_MPS', False) else DEVICE

class BaseEntailment:
    def save_prediction_cache(self):
        pass


class EntailmentLLM(BaseEntailment):

    def check_implication(self, text1, text2, question=None):
        if question is None:
            raise ValueError
        prompt = self.equivalence_prompt(text1, text2, question)

        logging.info('%s input: %s', self.name, prompt)

        response, logits, embeddings, = self.predict(prompt, temperature=0.02)

        logging.info('%s prediction: %s', self.name, response)

        binary_response = response.lower()[:30]
        if 'entailment' in binary_response:
            return 2
        elif 'neutral' in binary_response:
            return 1
        elif 'contradiction' in binary_response:
            return 0
        else:
            logging.warning('MANUAL NEUTRAL!')
            return 1


class EntailmentGPT4(EntailmentLLM):

    def __init__(self):
        super().__init__()
        self.name = 'gpt-4'
        self.client = AzureModel(model_name='gpt-4o')

    def equivalence_prompt(self, text1, text2, question):

        prompt = f"""We are evaluating answers to the question \"{question}\"\n"""
        prompt += "Here are two possible answers:\n"
        prompt += f"Possible Answer 1: {text1}\nPossible Answer 2: {text2}\n"
        prompt += "Does Possible Answer 1 semantically entail Possible Answer 2? Respond with entailment, contradiction, or neutral. Do not answer with anything else."""

        return prompt

    def predict(self, prompt, temperature):
        return self.client.predict(prompt, temperature)


class EntailmentGPT35(EntailmentGPT4):

    def __init__(self):
        super().__init__()
        self.name = 'gpt-3.5'


class EntailmentGPT4Turbo(EntailmentGPT4):

    def __init__(self):
        super().__init__()
        self.name = 'gpt-4-turbo'
        
class EntailmentGPT4o(EntailmentGPT4):

    def __init__(self):
        super().__init__()
        self.name = 'gpt-4o'


class EntailmentGPT4oMini(EntailmentGPT4):

    def __init__(self):
        super().__init__()
        self.name = 'gpt-4o-mini'


# class EntailmentLlama(EntailmentLLM):

#     def __init__(self, name):
#         super().__init__()
#         self.name = name
#         self.model = HuggingfaceModel(
#             name, stop_sequences='default', max_new_tokens=500)

#     def equivalence_prompt(self, text1, text2, question):

#         prompt = f"""We are evaluating answers to the prompt \"{question}\"\n"""
#         prompt += "Here are two possible answers:\n"
#         prompt += f"Possible Answer 1: {text1}\nPossible Answer 2: {text2}\n"
#         prompt += "Does Possible Answer 1 semantically entail Possible Answer 2? Respond only with entailment, contradiction, or neutral.\n"""
#         prompt += "Response:"""

#         return prompt

#     def predict(self, prompt, temperature):
#         predicted_answer, _, _ = self.model.predict(prompt, temperature)
#         return predicted_answer


def context_entails_response(context, responses, model):
    votes = []
    for response in responses:
        votes.append(model.check_implication(context, response))
    return 2 - np.mean(votes)


def get_semantic_ids(strings_list, model, strict_entailment=False, question=None):
    """Group list of predictions into semantic meaning."""

    def are_equivalent(text1, text2):

        implication_1 = model.check_implication(text1, text2, question)
        implication_2 = model.check_implication(text2, text1, question)  # pylint: disable=arguments-out-of-order
        assert (implication_1 in [0, 1, 2]) and (implication_2 in [0, 1, 2])

        if strict_entailment:
            semantically_equivalent = (implication_1 == 2) and (implication_2 == 2)

        else:
            implications = [implication_1, implication_2]
            # Check if none of the implications are 0 (contradiction) and not both of them are neutral.
            semantically_equivalent = (0 not in implications) and ([1, 1] != implications)

        return semantically_equivalent

    # Initialise all ids with -1.
    semantic_set_ids = [-1] * len(strings_list)
    # Keep track of current id.
    next_id = 0
    for i, string1 in enumerate(strings_list):
        # Check if string1 already has an id assigned.
        if semantic_set_ids[i] == -1:
            # If string1 has not been assigned an id, assign it next_id.
            semantic_set_ids[i] = next_id
            for j in range(i+1, len(strings_list)):
                # Search through all remaining strings. If they are equivalent to string1, assign them the same id.
                if are_equivalent(string1, strings_list[j]):
                    semantic_set_ids[j] = next_id
            next_id += 1

    assert -1 not in semantic_set_ids

    return semantic_set_ids


def logsumexp_by_id(semantic_ids, log_likelihoods, agg='sum_normalized'):
    """Sum probabilities with the same semantic id.

    Log-Sum-Exp because input and output probabilities in log space.
    """
    unique_ids = sorted(list(set(semantic_ids)))
    assert unique_ids == list(range(len(unique_ids)))
    log_likelihood_per_semantic_id = []

    for uid in unique_ids:
        # Find positions in `semantic_ids` which belong to the active `uid`.
        id_indices = [pos for pos, x in enumerate(semantic_ids) if x == uid]
        # Gather log likelihoods at these indices.
        id_log_likelihoods = [log_likelihoods[i] for i in id_indices]
        if agg == 'sum_normalized':
            # log_lik_norm = id_log_likelihoods - np.prod(log_likelihoods)
            log_lik_norm = id_log_likelihoods - np.log(np.sum(np.exp(log_likelihoods)))
            logsumexp_value = np.log(np.sum(np.exp(log_lik_norm)))
        else:
            raise ValueError
        log_likelihood_per_semantic_id.append(logsumexp_value)

    return log_likelihood_per_semantic_id


def predictive_entropy(log_probs):
    """Compute MC estimate of entropy.

    `E[-log p(x)] ~= -1/N sum_i log p(x_i)`, i.e. the average token likelihood.
    """

    entropy = -np.sum(log_probs) / len(log_probs)

    return entropy


def predictive_entropy_rao(log_probs):
    entropy = -np.sum(np.exp(log_probs) * log_probs)
    return entropy


def cluster_assignment_entropy(semantic_ids):
    """Estimate semantic uncertainty from how often different clusters get assigned.

    We estimate the categorical distribution over cluster assignments from the
    semantic ids. The uncertainty is then given by the entropy of that
    distribution. This estimate does not use token likelihoods, it relies soley
    on the cluster assignments. If probability mass is spread of between many
    clusters, entropy is larger. If probability mass is concentrated on a few
    clusters, entropy is small.

    Input:
        semantic_ids: List of semantic ids, e.g. [0, 1, 2, 1].
    Output:
        cluster_entropy: Entropy, e.g. (-p log p).sum() for p = [1/4, 2/4, 1/4].
    """

    n_generations = len(semantic_ids)
    counts = np.bincount(semantic_ids)
    probabilities = counts/n_generations
    assert np.isclose(probabilities.sum(), 1)
    entropy = - (probabilities * np.log(probabilities)).sum()
    return entropy
